"""Init Admin Page"""
import logging
import csv

from django.contrib import admin
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render

from articles.bgr_tasks import autoremove_task
from articles.tasks import read_all_media_files
from one_api.settings_prod import MEDIA_ROOT
from cropping.tasks import crop_up_news

from .forms import UserCreatorForm
from .models import Article, Medium, MediumType, ParentTag, Tag, Timeline, Overview, Files, User, UserActivityLogs

# Custom Actions for the admin site
from .post_migrations import assign_media_regionality, assign_media_to_generated_articles, execute_static_sql_statements, connect_tags_with_parents, create_standard_user
from .wizard.manipulators import create_users_in_db

logger = logging.getLogger(__name__)


class UserAdmin(admin.ModelAdmin):
    """ adding more actions to User in Admin sites """
    list_display = ['id_tag', 'activated']
    ordering = ['-activated']
    actions = ['deactivate_user', 'create_user', 'start_data_integration', 'apply_post_migrations', 'export_user_logs']

    @admin.action(description='Deactivate Users')
    def deactivate_user(self, request, queryset):
        """ instead of deleting we deactivate user """
        queryset.update(activated=False)
        self.message_user(request, f"Deactivated {queryset.count()} users")

    @admin.action(description='Create Users')
    def create_user(self, request, queryset):
        """ you can create a specific count of users """
        if 'apply' in request.POST:
            form = UserCreatorForm(request.POST)
            if form.is_valid():
                user_count_ = form.cleaned_data['user_count']
                create_users_in_db(user_count_)
                self.message_user(request, f"Created {user_count_} users")
            return HttpResponseRedirect(request.get_full_path())
        form = UserCreatorForm(initial={'user_count': 1})
        return render(request, 'admin/create_user.html', context={'form': form, 'items': queryset})

    @admin.action(description='Start Data Integration (PR + Crop)')
    def start_data_integration(self, request, _):
        """ starts parsing of PR files and cropping """
        # parse pressespiegel files
        read_all_media_files.delay()
        autoremove_task.delay(MEDIA_ROOT)
        # start cropper
        crop_up_news.delay()
        self.message_user(request, "Started data integration.")

    @admin.action(description='Apply post_migrations')
    def apply_post_migrations(self, request, _):
        """ manually applies all post_migrations that are also run on startup """
        execute_static_sql_statements()
        create_standard_user()
        connect_tags_with_parents()
        assign_media_regionality()
        assign_media_to_generated_articles()
        self.message_user(request, "Applied post migrations.")

    @admin.action(description='Download the user logs as .csv file')
    def export_user_logs(self, request, queryset):
        """Get logs for selected user(s) from database and export as csv"""
        filename = 'log_export'
        user_ids = []
        for user in queryset:
            user_ids.append(user.id_tag)
            filename += '_' + str(user.id_tag)
        logs = UserActivityLogs.objects.filter(user_id__in=user_ids).order_by('timestamp')
        if logs:
            response = download_csv(logs, filename + '.csv')
            return response

        self.message_user(request, "There are no logs for the selected user(s).")
        return None


class CustomTagFunctions(admin.ModelAdmin):
    """ adds custom admin action to articles.tag """
    actions = ['delete_unwanted_tags', 'move_cropping_categories']

    @admin.action(description='Delete all non pr tags')
    def delete_unwanted_tags(self, request, _):
        """ delete all non-pressespiegel tags """
        Tag.objects.exclude(parent__name="Pressespiegel").all().delete()
        self.message_user(request, "All unwanted tags deleted.")

    @admin.action(description='Move Tags from cropping')
    def move_cropping_categories(self, request, _):
        """ move cropping.categories to articles.tags """
        connect_tags_with_parents()
        self.message_user(request, "All tags moved.")


class CustomFilesFunctions(admin.ModelAdmin):
    """ adds custom admin action to articles.files """
    actions = ['delete_articles_by_file']

    @admin.action(description='Delete all articles for selected file')
    def delete_articles_by_file(self, request, queryset):
        """ start delete_articles_by_file from admin panel """
        article_count = Article.objects.filter(source_file__in=queryset).distinct().count()
        queryset.delete()
        self.message_user(request, f"The file and {article_count} articles were deleted.")


class UserActivityFunctions(admin.ModelAdmin):
    """ adds custom admin action to articles.useractivitylogs """
    actions = ['export_csv']

    @admin.action(description='Download all logs as .csv file')
    def export_csv(self, request, _):
        """Get all logs from database and export as csv"""
        logs = UserActivityLogs.objects.all().order_by('timestamp')
        if logs:
            response = download_csv(logs, 'log_export_all.csv')
            return response

        self.message_user(request, "There are no logs in the database.")
        return None


def download_csv(logs, filename):
    """Converts logs to CSV"""
    csvfile = HttpResponse(content_type='text/csv')
    csvfile['Content-Disposition'] = 'attachment; filename="' + filename + '"'

    # create the CSV writer
    writer = csv.writer(csvfile, delimiter=';')

    # get all model fields
    headers = [field.name for field in UserActivityLogs._meta.get_fields()]
    writer.writerow(headers)

    data = []
    for line in logs:
        row = []
        for field in headers:
            row.append(getattr(line, field, ''))
        data.append(row)

    # write all data rows to CSV
    writer.writerows(data)

    return csvfile


# Register your models here.
admin.site.register(Article)
admin.site.register(Medium)
admin.site.register(MediumType)
admin.site.register(Tag, CustomTagFunctions)
admin.site.register(ParentTag)
admin.site.register(Timeline)
admin.site.register(Overview)
admin.site.register(Files, CustomFilesFunctions)
admin.site.register(User, UserAdmin)  # Register Model with custom actions
admin.site.register(UserActivityLogs, UserActivityFunctions)
