""" All Tasks for articles """
import logging
import os
import datetime
import secrets

from celery import shared_task, chord
from django.utils import timezone
from django.db import connection

from articles.data_extraction.tasks import extract_information_from_file
from articles.models import Files, Article, Tag, TopMedia, Medium, MediumType, FunctionData, UserActivityLogs
from articles.serializers import union_medium_with_submedia
from cropping.models import Article as CroppingArticle
from one_api.settings_prod import MEDIA_ROOT, BASE_DIR

logger = logging.getLogger(__name__)


class CroppingArticles:
    """ Bundles the CroppingArticle QuerySet and its count to prevent DB calls in a loop """

    def __init__(self, articles, count):
        self.articles = articles
        self.count = count


class Media:
    """ Bundles the Medium QuerySet and its count to prevent DB calls in a loop """

    def __init__(self, top_media, distribution, count):
        self.top_media = top_media
        self.distribution = distribution
        self.count = count


@shared_task
def read_all_media_files():
    """
    Searches in the media folder for pdfs which has computable articles to save them in the database.
    """
    media_dir = os.listdir(MEDIA_ROOT)
    logger.info("read out %i files from media root", len([name for name in media_dir if name.endswith('pdf')]))
    celery_group = []
    for m_file in list(filter(lambda file_name: file_name.endswith('pdf'), media_dir)):
        if Files.objects.filter(name=m_file).count() == 0:
            file = Files(name=m_file, upload_date=timezone.now())
            file.save()
        elif Files.objects.filter(name=m_file, extraction_complete=False):
            logger.debug("file %s will be re-read", m_file)
            file = Files.objects.get(name=m_file)
            Article.objects.filter(source_file=file).delete()
            celery_group.append(extract_information_from_file.si(file.pk))
        else:
            logger.debug("file %s already exists and was correct extracted in db", m_file)
    chord(celery_group, compute_sql_statements.si('dynamic_tables.sql')).apply_async()


def generate_article(article_count, upload_date, tag_id, cropping_articles, media):
    """
        Generating article:
        article_count: amount of articles to generate with this call
        upload_date: defines the upload date of the articles that are to be created
        tag_id: id of the tag the article shall be associated with
        cropping_articles: related cropping articles
        media: related media
        First a list of articles with a length equal to article_count is created,
            containing similarly generated fake articles based on wizard_of_oz_faker.pdf file.
            After that, the articles get added to the through tag_id indexed tag.
    """
    file = Files.objects.get(name='wizard_of_oz_faker.pdf')

    objs = []
    for _ in range(article_count):
        if cropping_articles.count > 0:
            gen_title = cropping_articles.articles[secrets.randbelow(cropping_articles.count)].title
        else:
            gen_title = 'Fakeartikel'

        if media.count > 0:
            randint = secrets.randbelow(len(media.distribution))
            gen_medium = Medium.objects.get(name=media.distribution[randint])
        else:
            medium_type, _ = MediumType.objects.update_or_create(name='Zeitung')
            gen_medium, _ = Medium.objects.update_or_create(
                name=union_medium_with_submedia('fakemedium'),
                defaults={'medium_type': medium_type, 'reach': 1000}
            )

        input_data = {
            'date_of_publication': upload_date,
            'text': 'Dies ist ein beispielhafter Beispieltext',
            'pages': 2,
            'wrongly_parsed': False,
        }

        objs.append(Article(medium=gen_medium, source_file=file, name=gen_title, reach=gen_medium.reach, **input_data))

    tagged_obj = []
    tag = Tag.objects.get(id=tag_id)
    for obj in objs:
        tagged_obj.append(Article.tags.through(article=obj, tag=tag))
        obj.save()

    return tagged_obj


def generate_tag(tag_id):
    """
        Generating new fake tag:
        tag_id: suffix of the tag name, distiguishing multiple fake tags
        Creates a unique fake tag with given suffix.
    """
    tag = Tag(name="Fake" + str(tag_id))
    tag.save()


def create_fake_file_ine():
    """
        Adding fake file if not already exists:
        Checks, if file to relate fake articles to exists and fixes it's non-existance if necessary
    """
    if Files.objects.filter(name__exact='wizard_of_oz_faker.pdf').count() == 0:
        Files.objects.create(name='wizard_of_oz_faker.pdf', upload_date=datetime.datetime.today())


@shared_task
def generate_by_plot():
    """
        Generating fake articles by function graph:
        After checking if file to relate fake articles to exists,
            a fake article creation is started for all existing tags.
    """
    for tag in Tag.objects.exclude(name="account_used").exclude(parent__name="Pressespiegel").all():
        create_articles_per_tag.delay(tag.id)


@shared_task
def generate_by_id(tag_id):
    """
        Generating fake articles by tag id:
        tag_id: id of tag to generate the fake articles for
        After assuring the existance of file to relate articles to,
            a process to create fake articles for given tag is started.
    """
    create_fake_file_ine()
    create_articles_per_tag.delay(tag_id)


@shared_task
def create_articles_per_tag(tag_id):
    """
        Generating articles based on given tag:
        tag: tag the articles will be created for
        First, all the function data relevant to the given tag will be filtered from db,
            then this data will be skimmed through.
            Every instance of function data saves the amount of articles created for one day.
            For each day, the count of articles represented by the function is created.
            Finally, the generated articles get saved.
    """

    # setup
    cropping_articles = CroppingArticles(CroppingArticle.objects.all(), CroppingArticle.objects.all().count())
    top_medium = TopMedia.objects.exclude(name__contains='pressespiegel').all()
    top_media = Media(top_medium, create_distribution(top_medium), top_medium.count())
    tagged_objs = []
    start_date = get_start_date()
    tag = Tag.objects.get(id=tag_id)

    # get all the function values for given tag
    qdata = FunctionData.objects.filter(tag=tag)

    # generate articles based on y-axis values per day
    for qdat in qdata:
        tagged_objs += generate_article(qdat.y_data, start_date, tag.id, cropping_articles, top_media)
        start_date += datetime.timedelta(days=1)

    # insert generated articles into db
    Article.tags.through.objects.bulk_create(tagged_objs)


def get_start_date():
    """Defines and returns the start date for function, plot and article generation"""
    return datetime.date(2021, 12, 1)


@shared_task()
def compute_sql_statements(file_path):
    """
    Computes custom sql statements given in the file.
    Each statement should be finished with '# ---'.
    """
    etc_path = os.path.join('..', 'etc', file_path)
    with open(os.path.abspath(os.path.join(os.path.dirname(__file__), etc_path)), encoding='UTF-8') as file:
        sql_statement = file.read()
        with connection.cursor() as cursor:
            # TODO: replace check statement with something more robust
            for statement in sql_statement.split("# ---"):
                stripped_statement = statement.replace('\n', ' ').strip()
                cursor.execute(stripped_statement)
                row = cursor.fetchone()
            return row


def create_distribution(top_media):
    """ Creates a distribution based on the media article count """
    distribution = []
    for medium in top_media:
        for _ in range(medium.count):
            distribution.append(medium.name)

    return distribution


@shared_task
def log_to_db():
    """
    Task to extract user-dependent actions from the log and store them in the database.
    """
    with open(os.path.join(BASE_DIR, 'oneup.log'), 'r', encoding='UTF-8') as log_file:
        log_file_lines = log_file.readlines()
        log_file.close()

    logged_statements = 0
    if log_file_lines:
        for line in log_file_lines:
            if 'user=' in line:
                logged_statements += 1
                split_string = split_up_log_line(line)
                if split_string:
                    new_log = UserActivityLogs(
                        user_id=split_string[0],
                        action=split_string[1],
                        log_info=split_string[2],
                        timestamp=split_string[3]
                    )
                    new_log.save()
    return logged_statements


def split_up_log_line(line):
    """
    Splits the log lines into semantic units matching the UserActivityLogs database fields

    Lines are structured like this:
    2022-08-15 11:59:55,904 - INFO django.server.basehttp "GET /articles/top/media?mediaOverviewType=usertags&user=1 HTTP/1.1" 200 304
    """
    split_action = line.split('"')

    if len(split_action) > 1:
        split_timestamp = split_action[0].split(" - ")

        # e.g. GET /articles/top/media?mediaOverviewType=usertags&user=1 HTTP/1.1
        action = split_action[1]

        # e.g. 2022-08-15 11:59:55,904
        timestamp = datetime.datetime.strptime(split_timestamp[0].strip() + '000', '%Y-%m-%d %H:%M:%S,%f')
        timestamp = timezone.make_aware(timestamp)

        # e.g. INFO django.server.basehttp 200 304
        log_info = split_timestamp[1] + split_action[2].strip()

        split_user = action.split("user=")
        split_user = split_user[1].split("&")
        if not len(split_user) > 1:
            split_user = split_user[0].split(" ")

        # e.g. 1
        user = split_user[0]

        return [user, action, log_info, timestamp]

    return None
