""" Test file for pr_upload """
import logging
import tempfile
import datetime
from pathlib import Path
from os import listdir, mkdir, remove
from shutil import copy, rmtree
from unittest import mock

import pdfminer
from django.db.models.signals import post_save
from django.test import TransactionTestCase, TestCase

from articles.models import Article, Files, Tag, TopMedia, FunctionData
from articles.data_extraction.crop_meta import extract_infos
from articles.bgr_tasks import failed_upload, autoremove_task
from articles.tasks import CroppingArticles, Media, generate_tag, generate_article, create_articles_per_tag, \
    create_distribution
from articles.views import create_function

from cropping.models import Article as CroppingArticle

logger = logging.getLogger(__name__)


class CropPRIntegration(TransactionTestCase):
    """ Integration tests """

    def test_crop_file(self):
        """ tests the cropping algorithm in a whole file"""
        logger.info("started test_crop_file")
        dir_path = '/app/articles/data_extraction/'
        test_var = False
        has_pdf = False
        for file in listdir(dir_path):
            if "pdf" not in file:
                continue
            has_pdf = True
            try:
                with tempfile.TemporaryDirectory() as temp_dirname:
                    with mock.patch('articles.signals.extract_information_from_file',
                                    autospec=True) as mocked_post_save:
                        post_save.connect(mocked_post_save, sender=Files, dispatch_uid='extract_information_from_file')
                        path_file = dir_path + file
                        file = Files.objects.create(name=path_file, upload_date='18.11.2021')
                        extract_infos(file, temp_dirname, '')
                        test_var = True
            except pdfminer.pdfparser.PDFSyntaxError:
                self.fail("Couldn't parse PDF")
            break
        self.assertTrue(test_var & has_pdf)
        article = Article.objects.get(name='Offenes Fenster in Potsdam')
        self.assertEqual(article.reach, 187566)


class UploadView(TestCase):
    """ Test UploadView """

    def test_plain_view(self):
        """ test the view without any data """
        logger.info("started UploadView.test_plain_view")
        response = self.client.get('/pr_upload/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.context['documents']), 0)


MEDIA_ROOT = './media/'


def help_create_environment():
    """
        creates the environment for upload comparison
        this is necessary to have a consistent testing process:
            - purging data from previous test
            - creating vital folders
    """

    # removing data from previous tests
    ls_media = []
    try:
        ls_media = listdir(MEDIA_ROOT)  # getting directories from media folder
    except FileNotFoundError:
        pass
    if "testfile" in ls_media:  # removing previously gotten directories
        rmtree(MEDIA_ROOT)

    # creating folders for testfile and upload collections
    Path(MEDIA_ROOT).mkdir(parents=True)
    mkdir(MEDIA_ROOT + "testfile/")
    mkdir(MEDIA_ROOT + "upload/")


def rm_if_exists(target):
    """ removes a file if found in directory """
    if target in listdir("."):
        remove(target)


class BackgroundTasks(TransactionTestCase):
    """ Test the background tasks """

    def test_failed_upload(self):
        """
            Test if namebased failcheck for upload works:
            To test if the function registers a failed upload based on the namepattern,
            the test first adds two files with different names to the 'testfile' folder,
            then it adds a file with the same name as the first file to the 'upload' folder.
            expectation:
                the upload for the first file fails and succeeds for the second file
        """
        logger.info("started test_failed_upload")
        help_create_environment()  # cleaning the testing environment

        # filling the testfile folder
        copy("articles/Pressespiegel_20200602.pdf", MEDIA_ROOT + "testfile/Pressespiegel_20200702.pdf")
        copy("articles/Pressespiegel_20200602.pdf", MEDIA_ROOT + "testfile/Pressespiegel_20200703.pdf")

        # filling the upload folder
        copy("articles/Pressespiegel_20200602.pdf", MEDIA_ROOT + "upload/Pressespiegel_20200702.pdf")

        # check for naming duplicates
        uploads = listdir(MEDIA_ROOT + "upload")
        self.assertTrue(failed_upload("Pressespiegel_20200702.pdf", uploads))
        self.assertTrue(not failed_upload("Pressespiegel_20200703.pdf", uploads))
        # FIXME: is it necessary to fill testfile folder if i just check namepatterns hardcoded anyways

    def test_autoremove_task(self):
        """
            Test if autoremove task removes files according to his intended functionality:
            To test if the autoremove does not remove valid files,
                the test adds two valid files to the 'testfile' folder.
            To test the autoremove in case he encounters an invalid naming pattern,
                the test adds a wrongly named file to the 'testfile' folder. (*_a.pdf)
            To test the autoremove in case he encounters a already uploaded filename,
                the test adds a file that was previously uploaded
            expectation:
                Only the valid filenames are uploaded. Everything else gets removed.
                'testfile' is empty.
        """
        logger.info("started test_autoremove_task")
        help_create_environment()  # cleaning the testing environment

        # adding two valid and one invalid file
        copy("articles/Pressespiegel_20200602.pdf", MEDIA_ROOT + "testfile/")
        copy("articles/Pressespiegel_20200602.pdf", MEDIA_ROOT + "testfile/Pressespiegel_20200602_a.pdf")
        copy("articles/Pressespiegel_20200602.pdf", MEDIA_ROOT + "testfile/Pressespiegel_20200603.pdf")
        autoremove_task(MEDIA_ROOT)  # autoremove and upload

        # adding previously uploaded file
        copy("articles/Pressespiegel_20200602.pdf", MEDIA_ROOT + "testfile/Pressespiegel_20200603.pdf")
        autoremove_task(MEDIA_ROOT)  # autoremove and upload
        self.assertEqual(str(listdir(MEDIA_ROOT + "testfile")), "[]")
        self.assertEqual(str(listdir(MEDIA_ROOT + "upload")),
                         "['Pressespiegel_20200602.pdf', 'Pressespiegel_20200603.pdf']")


class FakeArticles(TransactionTestCase):
    """ Test the fake article creation """

    def test_create_fake_tag(self):
        """
            Test if fake tag creation for fake article creation works:
            Creating two new fake tags, then create a list of all existing tag names.
            Finally check if names of created tags are in the list of tag names.
            expectation:
                The tag names get created in pattern: "Fake"+<input_integer>,
                therefore the expected names will be found.
        """
        logger.info("started test_create_fake_tag")

        generate_tag(360)
        generate_tag(720)
        tags = [tag.name for tag in Tag.objects.all()]
        self.assertTrue("Fake360" in tags)
        self.assertTrue("Fake720" in tags)

    def test_create_fake_article(self):
        """
            Test if number of fake articles and a certain tag can be produced at a certain date:
            First a tag and file is created the new articles will be created for / referenced to,
                then the temporary article count is saved for later usage in conjunction with the check.
                Four articles will be created.
            Expectation: The delta article count after creation is four, having created four articles.
        """
        logger.info("started test_create_fake_article")

        # deal with setup tardiness
        with mock.patch('articles.signals.extract_information_from_file', autospec=True) as mocked_post_save:
            post_save.connect(mocked_post_save, sender=Files, dispatch_uid='extract_information_from_file')

            # generate tag to create articles for
            generate_tag(42)
            tag = Tag.objects.get(name="Fake42")

            # generate reference file for fake articles
            file = Files(name='wizard_of_oz_faker.pdf')
            file.save()

            # get article count before creation
            tcount = Article.objects.count()

            # get additional components
            cropping_articles = CroppingArticles(CroppingArticle.objects.all(), CroppingArticle.objects.all().count())
            top_medium = TopMedia.objects.all()
            top_media = Media(top_medium, create_distribution(top_medium), top_medium.count())

            # generate articles
            generate_article(4, datetime.date(1998, 10, 22), tag.id, cropping_articles, top_media)

            # compare with expectation
            self.assertEqual(Article.objects.count()-tcount, 4)

    def test_create_articles_per_tag(self):
        """
            Test if one specific tag is correctly provided with fake articles based on this tags plot:
            It starts off with a function generation for a new tag, used within this test.
                Then the "integral" of the function is found out, to represent the ideal amount of created articles.
                Furthermore, the articles get created according to the function, not the ideal amount.
                The amount of created articles gets calculated and compared with expectation.
            Expectation: The amount of created articles is identical to the ideal amount specified by the "integral".
            PS: I am very sorry for my usage of integral. It simply refers to the amount of all y values as a sum.
        """
        logger.info("started test_create_articles_per_tag")

        # setup tag
        generate_tag(84)
        ttag = Tag.objects.filter(name="Fake84")[0]
        create_function(ttag)

        # get ideal article amount
        y_data = FunctionData.objects.filter(tag=ttag)
        y_sum = sum(y.y_data for y in y_data)

        # create articles
        with mock.patch('articles.signals.extract_information_from_file', autospec=True) as mocked_post_save:
            post_save.connect(mocked_post_save, sender=Files, dispatch_uid='extract_information_from_file')
            file = Files(name='wizard_of_oz_faker.pdf')
            file.save()
            create_articles_per_tag(ttag.id)

            # get amount of created articles
            c_sum = len([x for x in Article.objects.all() if ttag.id in [y.id for y in x.tags.all()]])

        # check if assessment was correct
        self.assertEqual(y_sum, c_sum)
