""" Webcropper for uni-potsdam.de/nachrichten """
# TESTED 15th of November 2021
# !!RUNTIME 4h 19min
# pylint: disable=duplicate-code

import os
import logging
import time
import requests as rq
from bs4 import BeautifulSoup as bsx

from cropping.models import Category, Article, Meta
from one_api.settings_prod import APP_DATA_ROOT

logger = logging.getLogger(__name__)


def get_scoop(rq_url, rq_local):
    """
        Getting scoop for beautiful soup:
        rq_url: represents the url in question, can be web url or local file
        rq_local: true if local file is to be processed
        Requests and soups for either local or web pages.
    """

    # work with local file
    if rq_local:
        res = ""
        with open(rq_url, 'r', encoding='UTF-8') as page:
            res += page.read()
        return bsx(res, 'html.parser')

    # work with url, timeout if no response has been started after 30 sec
    res = rq.get(rq_url, timeout=30)
    return bsx(res.text, 'html.parser')


def get_categories(branch_prefix, branch_sets):
    """
        Getting all categories from uni-potsdam news page:
        branch_prefix: url, holding categories, until page index character
        branch_sets: url, holding categories, after page index character
        Gets all available categories by reading dropdown content on up news site.
    """
    # setup
    logger.info("starting the cropper")
    ls_curls = []

    # squeeze available categories
    logger.debug("squeeze available categories:")
    scoop = get_scoop(branch_prefix + str(1) + branch_sets + str(1), False)
    c_urls = list(map(lambda x: x['href'], scoop.findAll('a')))  # getting the category URLs
    c_urls = list(filter(lambda x: x.startswith("/de/nachrichten/page?tx_news_pi"), c_urls))
    c_names = list(filter(lambda x: x.text.startswith(" "), scoop.findAll('a')))  # getting names
    for url, name in zip(c_urls, c_names):
        ls_curls.append((url[70:], name.text[1: len(name.text) - 1]))
        cat_name = str(name.text[1: len(name.text) - 1])
        cat, created = Category.objects.update_or_create(name=cat_name)
        if created:
            logger.debug('saved category %s (%s)', cat.name, cat.pk)

    logger.debug(str(len(ls_curls)) + " categories found")
    return ls_curls


def run_crop(categories, branch, art_local, crop_session, meta):
    """
        Let the cropper process all categories:
        categories: list holding all categories names with their urls as tuple (url,name)
        branch: holding tuple of what later becomes (branch_prefix,branch_sets) in get_categories
        art_local: true if article is a local file, not referenced as url
        crop_session: related session
        meta: data that saves progression of running cropping process, should it end abruptly
            cropper will automatically continue cropping process where it last left off if so desired
        Method iterates through all given categories and extracts their article urls.
            After extracting all articles of a category the found articles are read.
    """
    for dv_url in categories[meta.cath_index:]:
        logger.debug("gathering on topic: " + dv_url[1])
        ls_urls, branch_page = [], meta.page_index

        # getting relevant URLs
        extract_relevant_urls(art_local, branch, branch_page, dv_url, ls_urls)

        # work with found articles
        logger.debug("starting to read the found articles")
        read_articles(ls_urls, dv_url, art_local, crop_session, meta)

        # prepare meta for next category
        meta.cath_index += 1
        meta.page_index = 1
        meta.art_index = 0


def extract_relevant_urls(art_local, branch, branch_page, dv_url, ls_urls):
    """
        Extracting relevant urls from a html site:
        art_local: holds true as value, if local article is to be processed
        branch: holds tuple of (branch_prefix,branch_sets)
        branch_page: page index of url finder in current category
        dv_url: url of current category
        ls_urls: url collection of all found articles (originally empty list), to be modified by method
        Method iterates through all pages of the current category.
            As long as a page has articles all article urls get saved and the page number increases.
            If a page has no articles, the last page has been reached and the method terminates.
    """
    while True:
        logger.debug("getting available links on page " + str(branch_page))
        logger.info(dv_url[0])

        # create url
        branch_url = branch[0] + str(branch_page) + branch[1] + dv_url[0]

        # find article urls on page
        if art_local:
            scoop = get_scoop(dv_url[0], True)
            t_urls = map(str, scoop.findAll('a'))
            t_urls = list(filter(lambda x: x.startswith("<a href"), t_urls))
            t_urls = list(filter(lambda x: "html" in x, t_urls))
            t_urls = list(map(lambda x: x.split('href="', 1)[1], t_urls))
            t_urls = list(map(lambda x: x.split('"', 1)[0], t_urls))
            t_urls = list(filter(lambda x: x.startswith("/"), t_urls))
            t_urls = list(dict.fromkeys(t_urls))
            url_prefix = os.getcwd()
        else:
            scoop = get_scoop(branch_url, False)
            t_urls = list(map(lambda x: x['href'], scoop.findAll('a')))
            t_urls = list(filter(lambda x: x.startswith("/de/nachrichten/detail"), t_urls))
            url_prefix = 'https://www.uni-potsdam.de'

        # terminate if no urls exist on page
        if not t_urls:
            break

        # filter immediate url duplicates and add prefix
        t_last = ""
        for t_url in t_urls:
            if t_last != t_url:
                ls_urls.append(url_prefix + t_url)
            t_last = t_url

        # increase page number or terminate if page is local
        branch_page += 1
        if art_local:
            break


def read_articles(articles, category, art_local, crop_session, meta):
    """
        Read and process articles:
        articles: list of previously found article urls to read
        category: tuple of current category. pattern -> (category_url,category_name)
        art_local, crop_session, meta: same function as in run_crop
        Method reads all given urls in articles list and writes their contents to export_raw.csv.
            Also, after every write, the meta information gets updated in case the cropper is stopped.
    """
    for url in get_url(articles, meta):
        logger.info("now reading: " + url)

        # read contents of article
        s_article = crop_html(art_local, category, crop_session, meta, url)

        # write to csv
        write_categories_to_file(category, s_article)

        # update meta information
        meta.art_index += 1
        if meta.art_index > 9:  # category url can hold no more than 10 articles
            meta.page_index += 1
            meta.art_index = 0
        meta.save()


def write_categories_to_file(category, s_article):  # FIXME: naming and misconceptions
    """
        Writes the articles alongside their respective categories into export_raw.csv:
        category: tuple of (category_url,category_name) !!can be changed to only hold the name!!
        s_article: string of article contents to write to file
        !!change name later, not only categories get written!!
    """
    with open(APP_DATA_ROOT + "export_raw.csv", "a", encoding='UTF-8') as tch:
        tch.write(category[1] + "," + s_article + "\n")


def get_url(articles, meta):
    """
        Get url from list by meta information:
        articles: list of articles
        meta: meta information
    """
    return articles[meta.art_index:]


def crop_html(art_local, category, crop_session, meta, url):
    """
        Extract title and relevant text from html:
        art_local, category, crop_session, meta: see read_articles
        url: url to read article from
    """
    # respectfully waiting if necessary
    if meta.wait_index > 200:
        logger.debug("respectfully waiting for 1 minutes...")
        time.sleep(60 * int(not art_local))
        meta.wait_index = 0
    else:
        meta.wait_index += 1
    # compiling article
    scoop, s_article = get_scoop(url, art_local), ""
    up_content = scoop.find(id='up_content')
    title = up_content.find('h1')
    paragraphs = up_content.findAll('p')
    for paragraph in paragraphs:
        s_article += " ".join(paragraph.text.split()) + "\n"
    title_text = " ".join(title.text.strip().split())
    article, _ = Article.objects.update_or_create(
        title=title_text,
        defaults={'text': s_article, 'session_id': crop_session.id}
    )
    logger.debug("article saved: %i", article.id)
    s_article = title_text + "\n\n" + s_article
    cat_obj, _ = Category.objects.update_or_create(name=str(category[1]))
    logger.debug("added category %s to article %i", cat_obj.name, article.id)
    article.categories.add(cat_obj)
    article.save()
    return s_article


def join_categories():
    """
        Join duplicated articles from different categories:
        The joiner reads through the raw export file export_raw.csv, searching for article duplications.
            An article, that does not exist, gets written to a string, collecting all articles.
            If an article has already been written previously, only it's category gets added to the
            existing instance of the article in the string.
            After the eof has been reached the string will be written to export.csv.
    """
    logger.info("searching & joining duplicates...")

    # build collection string and filter duplicates
    with open(APP_DATA_ROOT + "export_raw.csv", "r", encoding='UTF-8') as rfile:
        lines = list(map(lambda x: x.split(',', 1), rfile.readlines()))
        join_categories_in_lines(lines)

    # rewrite joined information as updated csv
    logger.info("writing the final .csv file...")
    with open("export.csv", "a", encoding='UTF-8') as rwr:
        for line in lines:
            rwr.write(line[0] + ',' + line[1])


def join_categories_in_lines(lines):
    """
        Join duplicated categories in line:
        lines: existing collection string containing all articles (to be modified)
        All articles in the string get processed one by one.
            From the current article on, every article that follows gets compared.
            Should the articles be identical, the second article gets removed but it's category gets added
            to the original line holding the article in question.
    """
    for j, line in enumerate(lines):
        i = j + 1
        while i < len(lines):
            if line[1] == lines[i][1]:
                logger.debug("joining duplicate between \"" + line[0] + "\" and \"" + lines[i][0] + "\"")
                line[0] += ';' + lines[i][0]
                lines.pop(i)
            else:
                i += 1


def start_crop(prefix, sets, crop_session, restart):
    """
        Runs the entire cropping process:
        prefix, sets: see get_categories
        crop_session: related session
        restart: holds boolean value if cropper should restart and ignore saved cropping progress
        First, the progression data from the last cropping process is loaded, to finish where the cropper last left off.
            Then, if the progression data did not enable the joining phase, the article cropping proceeds.
            Finally, when in join phase, the categories of duplicate articles get joined in a new export file.
    """

    # load last cropping process data
    if Meta.objects.count() > 0 and not restart:
        meta = Meta.objects.all()[0]
    else:
        meta = Meta()
        meta.save()

    # run file cropping if not yet in joining phase
    if not meta.is_joining:
        run_crop(get_categories(prefix, sets), (prefix, sets), False, crop_session, meta)
        meta.is_joining = True
        meta.save()

    join_categories()
    meta.is_joining = False
    meta.save()
