""" Test class """

import logging
import os.path
from unittest.mock import mock_open, patch, call

from django.utils import timezone

from cropping.models import CropSession, Article, Meta
from cropping.cropper import join_categories_in_lines, crop_html, get_url, extract_relevant_urls, run_crop
from cropping.cropper import get_categories, get_scoop
from one_api.settings_prod import APP_DATA_ROOT
from articles.test_with_articles import CreateArticlesWithTagsTestCase

logger = logging.getLogger(__name__)

BRANCH = 'https://www.uni-potsdam.de/de/nachrichten/page-'
SETS = '?tx_news_pi1[overwriteDemand][categories]='
TESTFILES_DIR = os.path.abspath('.') + '/cropping/testfiles'


class WebMediaCropper(CreateArticlesWithTagsTestCase):
    """ Test the mediacropping of webarticles """

    def test_get_scoop(self):
        """
            Test if beautiful soup scoop was produced:
            To proceed with the cropping process, the cropper needs a return on get_scoop,
            additionally this test secretly serves as a check, if the webpage to crop from even exists.
            expectation:
                A return exists.
        """
        logger.info("started WebMediaCropper.test_get_scoop")
        scoop = get_scoop("https://www.uni-potsdam.de/de/nachrichten/", False)
        self.assertTrue(scoop is not None)

    def test_get_categories(self):
        """
            Test if all anticipated categories have been found:
            This test does not check if all categories have been found, this is an overcomplication,
            it tests if any categories have been found and derives from the solution if the function works.
            The process basically is:
                If any categories have been found, the function found all existing categories.
            expectation:
                Cropper will read the page correctly and therefore find articles.
        """
        logger.info("started WebMediaCropper.test_get_categories")
        categories = get_categories(BRANCH, SETS)
        self.assertTrue(len(categories) > 0)

    def test_run_crop(self):
        """
            Test if the cropping has been successful:
            To test if the cropper writes the correct information to the raw csv file,
            a predictable html has to be cropped and the output has to be compared to a correct csv file.
            expectation:
                The written export_raw.csv file is identical to the given example file.
        """
        logger.info("started WebMediaCropper.test_run_crop")
        categories = [(f'{TESTFILES_DIR}/articles.html', 'Ethik')]
        crop_session = CropSession(date_of_crop=timezone.now())
        crop_session.save()
        meta = Meta()
        meta.save()
        input_data = ""
        # enter input data because the mocking includes also open-read functions
        # and if run_crop doesn't find any text to compute the algorithm would stop
        with open(f'{TESTFILES_DIR}/articles.html', 'r', encoding='UTF-8') as page:
            input_data += page.read()
        self.assertTrue(len(input_data) > 0)
        mocked_open_functions = mock_open(read_data=input_data)
        with patch('builtins.open', mocked_open_functions):
            run_crop(categories, (BRANCH, SETS), True, crop_session, meta)
        absolut_path = os.path.abspath('.')
        mocked_calls = [
            call(absolut_path + '/cropping/testfiles/articles.html', 'r', encoding='UTF-8'),
            call(absolut_path + '/cropping/testfiles/alpen.html', 'r', encoding='UTF-8'),
            call(APP_DATA_ROOT + 'export_raw.csv', 'a', encoding='UTF-8'),
            call().write('Ethik,Neues aus der Universität Potsdam\n\nalle anzeigen\n\n')
        ]
        mocked_open_functions.assert_has_calls(mocked_calls, any_order=True)

    def test_extract_relevant_urls(self):
        """ test if the cropping has been successful """
        logger.info("started WebMediaCropper.test_extract_relevant_urls")
        categories = [(f'{TESTFILES_DIR}/articles.html', 'Ethik')]
        crop_session = CropSession(date_of_crop=timezone.now())
        crop_session.save()
        meta = Meta()
        meta.save()
        meta_cat = categories[meta.cath_index:]
        ls_urls = []
        extract_relevant_urls(crop_session, (BRANCH, SETS), meta.page_index, meta_cat[0], ls_urls)
        self.assertTrue(len(ls_urls) == 2)
        self.assertTrue([s for s in ls_urls if 'alpen.html' in s])
        self.assertTrue([s for s in ls_urls if 'lesfemmes.html' in s])

    def test_read_articles(self):
        """
            Test if articles have been read correctly:
            This tests only the article reader the run_crop method uses, to read all articles.
            If test_run_crop was successful, this test will probably succeed as well.
            The reader will be tested just like the run_crop.
            The output will be compared to an archetypical file and valued depending on the comparison.
            expectation:
                The export_raw.csv will be identical to the archetypical test_result.csv.
        """
        logger.info("started WebMediaCropper.test_read_articles")
        articles = [f"{TESTFILES_DIR}/test_alpen.html"]
        crop_session = CropSession(date_of_crop=timezone.now())
        crop_session.save()
        meta = Meta()
        meta.save()

        cropped_html = crop_html(True, ("null", "Ethik"), crop_session, meta, get_url(articles, meta)[0])
        self.assertEqual(meta.wait_index, 1)
        self.assertEqual("Unterwegs in den Alpen: 6. Juni 2018\n\nello another\n", cropped_html)
        article = Article.objects.all()
        self.assertEqual(article[0].title, "Unterwegs in den Alpen: 6. Juni 2018")
        self.assertEqual(article[0].text, "ello another\n")

    def test_join_categories(self):
        """
            Test if all duplicated categories have been joined:
            One article can be found in multiple different categories. This results in duplicates!
            To prevent that, the export_raw.csv is further processed, resulting in export.csv.
            This processing joins the duplicates and lists them once, alongside all their categories.
            To test this function we write a fake export_raw.csv with multiple duplicates,
            then, after joining, we compare the export.csv to the expected output pattern.
            expectation:
                The expected output pattern is identical to the export.csv file contents.
        """
        logger.info("started WebMediaCropper.test_join_categories")
        all_lines = ["Sabretoothtigerism,Once upon a time\n", "Woolspinnerism,Happily ever after\n",
                     "Lannisterism,Happily ever after\n", "Loneliness,I will be alone\n",
                     "Goblin Studies,Once upon a time\n"]
        lines = list(map(lambda x: x.split(',', 1), all_lines))
        join_categories_in_lines(lines)
        target = [["Sabretoothtigerism;Goblin Studies", "Once upon a time\n"],
                  ["Woolspinnerism;Lannisterism", "Happily ever after\n"],
                  ["Loneliness", "I will be alone\n"]]
        self.assertListEqual(target, lines)
